"""The Ubuntu One Music Store Rhythmbox plugin."""
# Copyright (C) 2009 Canonical, Ltd.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License
# version 3.0 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License version 3.0 for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.
#
# Authored by Stuart Langridge <stuart.langridge@canonical.com>

 
import gtk, gobject, os, urllib, gconf
import gst, gst.pbutils
import aptdaemon.client
from aptdaemon.enums import *
from aptdaemon.gtkwidgets import AptErrorDialog, \
                                 AptProgressBar
import rb, rhythmdb
try:
    from ubuntuone.gtkwidgets import MusicStore as U1MusicStore
    using_prototype = False
except ImportError:
    from pythonprototype import U1MusicStorePythonPrototype as U1MusicStore
    using_prototype = True
    
import gettext
from gettext import lgettext as _
gettext.bindtextdomain("rhythmbox-ubuntuone-music-store", "/usr/share/locale")
gettext.textdomain("rhythmbox-ubuntuone-music-store")

U1LIBRARYPATH = "file://%s" % os.path.expanduser("~/.ubuntuone/musicstore")
RB_LIBRARY_LOCATIONS = "/apps/rhythmbox/library_locations"

class U1MusicStorePlugin (rb.Plugin):
    """The Ubuntu One Music Store."""

    def __init__(self):
        rb.Plugin.__init__(self)

    def activate(self, shell):
        """Plugin startup."""
        self.db = shell.get_property("db")
        group = rb.rb_source_group_get_by_name ("stores")
        if not group:
            group = rb.rb_source_group_register ("stores",
                                           "Stores",
                                           rb.SOURCE_GROUP_CATEGORY_FIXED)
        theme = gtk.icon_theme_get_default()
        rb.append_plugin_source_path(theme, "/icons")
        width, height = gtk.icon_size_lookup(gtk.ICON_SIZE_LARGE_TOOLBAR)
        icon = rb.try_load_icon(theme, "ubuntuone-client-working", width, 0)
        self.entry_type = self.db.entry_register_type("U1EntryType")
        self.source = gobject.new (U1Source,
                                   shell=shell,
                                   entry_type=self.entry_type,
                                   source_group=group,
                                   icon=icon,
                                   plugin=self)
        shell.register_entry_type_for_source(self.source, self.entry_type)
        shell.append_source(self.source, None) # Add the source to the list
        self.shell = shell
        self.source.connect("preview-mp3", self.play_preview_mp3)
        self.add_u1_library()

    def deactivate(self, shell):
        """Plugin shutdown."""
        pass

    def play_preview_mp3(self, source, url, title):
        """Play a passed mp3; signal handler for preview-mp3 signal."""
        # create an entry, don't save it, and play it
        entry = self.shell.props.db.entry_lookup_by_location(url)
        if entry is None:
            entry = self.shell.props.db.entry_new(self.entry_type, url)
        self.shell.props.db.set(entry, rhythmdb.PROP_TITLE, title)
        player = self.shell.get_player()
        player.stop()
        player.play()
        player.play_entry(entry, self.source)
        # FIXME delete this entry when it finishes playing. Don't know how yet.

    def add_u1_library(self):
        """Add the U1 library if not listed in RB and re-add if changed."""
        client = gconf.client_get_default()
        self._add_u1_library_if_not_present(client)
        # Watch for changes to the gconf key and re-add the library
        client.notify_add(RB_LIBRARY_LOCATIONS, 
            self._add_u1_library_if_not_present)

    def _add_u1_library_if_not_present(self, client, *args, **kwargs):
        """Check for the U1 library and add to libraries list."""
        libraries = client.get_list(RB_LIBRARY_LOCATIONS, gconf.VALUE_STRING)
        if U1LIBRARYPATH not in libraries:
            libraries.append(U1LIBRARYPATH)
            client.set_list(RB_LIBRARY_LOCATIONS, gconf.VALUE_STRING, libraries)

class U1Source(rb.Source):
    """A Rhythmbox source widget for the U1 Music Store."""
    # gproperties required so that rb.Source is instantiable
    __gproperties__ = {
      'plugin': (rb.Plugin, 'plugin', 'plugin', 
                    gobject.PARAM_WRITABLE|gobject.PARAM_CONSTRUCT_ONLY),
    }
    # we have the preview-mp3 signal; we receive it from the widget, and re-emit
    # it so that the plugin gets it, because the plugin actually plays the mp3
    __gsignals__ = {
      "preview-mp3": (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, (str, str)),
    }
    
    def __init__(self):
        rb.Source.__init__(self, name=_("Ubuntu One"))
        self.__activated = False

    def do_impl_activate(self):
        """Source startup."""
        if self.__activated: return
        self.__activated = True
        self.body = gtk.ScrolledWindow()
        self.pack_start(self.body)
        self.body.show_all()
        self.test_can_play_mp3()
        rb.Source.do_impl_activate (self)

    def test_can_play_mp3(self):
        """Can the user play mp3s? Start a GStreamer pipeline to check."""
        mp3pth = os.path.realpath(os.path.join(
            os.path.split(__file__)[0], "empty.mp3"))
        uri = "file://%s" % urllib.quote("%s" % mp3pth)
        self.install_pipeline = gst.parse_launch(
          'uridecodebin uri=%s ! fakesink' % uri)
        bus = self.install_pipeline.get_bus()
        bus.add_signal_watch()
        bus.connect("message::element", self._got_element_msg)
        bus.connect("message::eos", self._got_end_of_stream)
        self.install_pipeline.set_state(gst.STATE_PLAYING)

    def _got_element_msg(self, bus, msg):
        """Handler for element messages from the check-mp3 pipeline.
           GStreamer throws a "plugin-missing" element message if the
           user does not have the right codecs to play a file."""
        plugin_missing = gst.pbutils.is_missing_plugin_message(msg)
        if plugin_missing:
            self.install_pipeline.set_state(gst.STATE_NULL)
            self.install_mp3_playback()

    def _got_end_of_stream(self, bus, msg):
        """Handler for end of stream from the check-mp3 pipeline.
           If we reach the end of the stream, mp3 playback is enabled."""
        self.install_pipeline.set_state(gst.STATE_NULL)
        self.add_music_store_widget()

    def install_mp3_playback(self):
        """Use aptdaemon to install the Fluendo mp3 playback codec package."""
        self.install_box = gtk.Alignment(xscale=0.0, yscale=0.0, xalign=0.5,
            yalign=0.5)
        self.install_vbox = gtk.VBox()
        self.install_label_head = gtk.Label()
        self.install_label_head.set_use_markup(True)
        not_installed = _("MP3 plugins are not installed")
        self.install_label_head.set_markup('<span weight="bold" size="larger">'
            '%s</span>' % not_installed)
        self.install_label_head.set_alignment(0.0, 0.5)
        self.install_label_body = gtk.Label()
        self.install_label_body.set_text(_('To listen to your purchased songs'
            ', you need to install MP3 plugins. Click below to install them.'))
        self.install_label_body.set_alignment(0.0, 0.5)
        self.install_hbtn = gtk.HButtonBox()
        self.install_hbtn.set_layout(gtk.BUTTONBOX_END)
        self.install_button = gtk.Button(label=_("Install MP3 plugins"))
        self.install_button.connect("clicked", self._start_mp3_install)
        self.install_hbtn.add(self.install_button)
        self.install_vbox.pack_start(self.install_label_head, expand=False)
        self.install_vbox.pack_start(self.install_label_body, expand=False,
            padding=12)
        self.install_vbox.pack_start(self.install_hbtn, expand=False)
        self.install_box.add(self.install_vbox)
        self.install_box.show_all()
        self.body.add_with_viewport(self.install_box)

    def _start_mp3_install(self, btn):
        ac = aptdaemon.client.AptClient()
        ac.install_packages(["gstreamer0.10-fluendo-mp3"],
                        reply_handler=self._run_transaction,
                        error_handler=self._on_error)

    def _run_transaction(self, transaction):
        """Show progress of aptdaemon package installation."""
        self.install_progress = AptProgressBar(transaction)
        self.install_progress.show()
        self.install_label_head.set_text("")
        self.install_label_body.set_text("Installing MP3 plugins")
        self.install_hbtn.hide()
        self.install_vbox.pack_start(self.install_progress, expand=False)
        transaction.run(reply_handler=lambda: True, 
                        error_handler=self._on_error)
        transaction.connect("finished", self._finished)

    def _finished(self, trans, exit_code):
        """Aptdaemon package installation finished; show music store."""
        if exit_code == 0:
            self.body.remove(self.install_box.get_parent())
            gst.update_registry()
            self.add_music_store_widget()
        else:
            self._on_error("Could not find the "
                "gstreamer0.10-fluendo-mp3 package.")

    def _on_error(self, error):
        """Error handler for aptdaemon."""
        print error
        problem_installing = _("There was a problem installing, sorry")
        self.install_label_head.set_markup('<span weight="bold" size="larger">'
            '%s</span>' % problem_installing)
        self.install_label_body.set_text(_('Check your internet connection and '
            'try again.'))
        if getattr(self, "install_progress"):
            self.install_progress.hide()
        self.install_hbtn.show()

    def add_music_store_widget(self):
        """Display the music store widget in Rhythmbox."""
        self.browser = U1MusicStore()
        if using_prototype:
            storeurl = os.environ.get("U1MUSICSTOREURL", None)
            vbox = gtk.VBox()
            if storeurl:
                labeltext = ("WARNING: You are using the Python "
                    "prototype of the music store widget!")
            else:
                labeltext = ("WARNING: You are using the Python "
                    "prototype of the music store widget; it is assumed that "
                    "you're doing this for testing, but you have not "
                    "specified U1MUSICSTOREURL on the command line. Re-run "
                    "Rhythmbox as "
                    "'U1MUSICSTOREURL=http://localhost:xxxxx/music/store "
                    "rhythmbox -D U1'.")
            lbl = gtk.Label(labeltext)
            lbl.set_line_wrap(True)
            vbox.pack_start(lbl, expand=False, fill=False)
            vbox.pack_start(self.browser, expand=True, fill=True)
            self.body.add(vbox)
        else:
            self.body.add(self.browser)
        self.body.show_all()
        self.browser.set_no_show_all(True)
        self.browser.set_property("visible", True)
        self.browser.connect("preview-mp3", self.re_emit_preview)

    def do_impl_can_pause(self): 
        """Implementation can pause.
           If we don't handle this, Rhythmbox segfaults."""
        return True # so we can pause, else we segfault

    def re_emit_preview(self, widget, url, title):
        """Handle the preview-mp3 signal and re-emit it to the rb.Plugin."""
        self.emit("preview-mp3", url, title)

    def do_set_property(self, property, value):
        """Allow property settings to handle the plugin call."""
        if property.name == 'plugin':
            self.__plugin = value
        else:
            raise AttributeError, 'unknown property %s' % property.name


